import pytest

from workflow_engine import (
    Edge,
    FloatValue,
    SequenceValue,
    StringMapValue,
    ValueSchemaValue,
    Workflow,
)
from workflow_engine.contexts import InMemoryContext
from workflow_engine.core.io import InputNode, OutputNode, SchemaParams
from workflow_engine.core.values.schema import SequenceValueSchema
from workflow_engine.execution import TopologicalExecutionAlgorithm
from workflow_engine.nodes import AddNode, ForEachNode


@pytest.fixture
def add_workflow() -> Workflow:
    """Create a workflow that adds two numbers: a + b = c."""
    input_node = InputNode(
        id="input",
        params=SchemaParams(
            fields=StringMapValue[ValueSchemaValue](
                {
                    "a": ValueSchemaValue(FloatValue.to_value_schema()),
                    "b": ValueSchemaValue(FloatValue.to_value_schema()),
                }
            )
        ),
    )
    output_node = OutputNode(
        id="output",
        params=SchemaParams(
            fields=StringMapValue[ValueSchemaValue](
                {
                    "c": ValueSchemaValue(FloatValue.to_value_schema()),
                }
            )
        ),
    )

    add = AddNode(id="add")

    return Workflow(
        input_node=input_node,
        output_node=output_node,
        inner_nodes=[add],
        edges=[
            Edge.from_nodes(
                source=input_node,
                source_key="a",
                target=add,
                target_key="a",
            ),
            Edge.from_nodes(
                source=input_node,
                source_key="b",
                target=add,
                target_key="b",
            ),
            Edge.from_nodes(
                source=add,
                source_key="sum",
                target=output_node,
                target_key="c",
            ),
        ],
    )


@pytest.fixture
def workflow(add_workflow: Workflow) -> Workflow:
    input_node = InputNode(
        id="input",
        params=SchemaParams(
            fields=StringMapValue[ValueSchemaValue](
                {
                    "sequence": ValueSchemaValue(
                        SequenceValueSchema(
                            type="array",
                            items=add_workflow.input_schema,
                        )
                    ),
                }
            )
        ),
    )
    output_node = OutputNode(
        id="output",
        params=SchemaParams(
            fields=StringMapValue[ValueSchemaValue](
                {
                    "results": ValueSchemaValue(
                        SequenceValueSchema(
                            type="array",
                            items=add_workflow.output_schema,
                        )
                    ),
                }
            )
        ),
    )

    for_each = ForEachNode.from_workflow(
        id="for_each",
        workflow=add_workflow,
    )

    workflow = Workflow(
        input_node=input_node,
        output_node=output_node,
        inner_nodes=[for_each],
        edges=[
            Edge.from_nodes(
                source=input_node,
                source_key="sequence",
                target=for_each,
                target_key="sequence",
            ),
            Edge.from_nodes(
                source=for_each,
                source_key="sequence",
                target=output_node,
                target_key="results",
            ),
        ],
    )

    return workflow


@pytest.mark.asyncio
async def test_for_each_simple_sequence(workflow: Workflow):
    """Test that ForEachNode processes a simple sequence of addition operations."""
    context = InMemoryContext()
    algorithm = TopologicalExecutionAlgorithm()

    input = workflow.input_type.model_validate(
        {
            "sequence": [
                {"a": 1.0, "b": 2.0},
                {"a": 3.0, "b": 4.0},
                {"a": 5.0, "b": 6.0},
            ]
        }
    ).to_dict()

    errors, output = await algorithm.execute(
        context=context,
        workflow=workflow,
        input=input,
    )

    assert not errors.any(), errors

    # Compare actual values (dynamically generated DataValue types have identity issues)
    results = output["results"]
    assert isinstance(results, SequenceValue)
    assert len(results) == 3
    assert results[0].root.c == FloatValue(3.0)
    assert results[1].root.c == FloatValue(7.0)
    assert results[2].root.c == FloatValue(11.0)


@pytest.mark.asyncio
async def test_for_each_empty(workflow: Workflow):
    """Test that ForEachNode processes an empty sequence."""
    context = InMemoryContext()
    algorithm = TopologicalExecutionAlgorithm()

    input = workflow.input_type.model_validate(
        {
            "sequence": [],
        }
    ).to_dict()

    errors, output = await algorithm.execute(
        context=context,
        workflow=workflow,
        input=input,
    )

    assert not errors.any(), errors

    results = output["results"]
    assert isinstance(results, SequenceValue)
    assert len(results) == 0



# workflow_engine/core/__init__.py
from .context import Context
from .edge import Edge
from .engine import WorkflowEngine
from .error import NodeException, ShouldRetry, UserException, WorkflowErrors
from .execution import ExecutionAlgorithm
from .io import InputNode, OutputNode
from .migration import (
    Migration,
    MigrationError,
    MigrationNotFoundError,
    MigrationRegistry,
    MigrationRunner,
    MigrationValidationError,
    clean_edges_after_migration,
    load_workflow_with_migration,
    migration_registry,
    migration_runner,
)
from .node import (
    Empty,
    Node,
    NodeRegistry,
    NodeTypeInfo,
    Params,
)
from .values import (
    JSON,
    BooleanValue,
    Caster,
    Data,
    DataMapping,
    DataValue,
    File,
    FileValue,
    FloatValue,
    IntegerValue,
    JSONValue,
    NullValue,
    SequenceValue,
    StringMapValue,
    StringValue,
    Value,
    ValueRegistry,
    ValueSchema,
    ValueSchemaValue,
    ValueType,
)
from .workflow import Workflow, WorkflowValue

__all__ = [
    "BooleanValue",
    "Caster",
    "clean_edges_after_migration",
    "Context",
    "Data",
    "DataMapping",
    "DataValue",
    "Edge",
    "Empty",
    "ExecutionAlgorithm",
    "File",
    "FileValue",
    "FloatValue",
    "InputNode",
    "IntegerValue",
    "JSON",
    "JSONValue",
    "load_workflow_with_migration",
    "migration_registry",
    "migration_runner",
    "Migration",
    "MigrationError",
    "MigrationNotFoundError",
    "MigrationRegistry",
    "MigrationRunner",
    "MigrationValidationError",
    "Node",
    "NodeException",
    "NodeRegistry",
    "NodeTypeInfo",
    "NullValue",
    "OutputNode",
    "Params",
    "SequenceValue",
    "ShouldRetry",
    "StringMapValue",
    "StringValue",
    "UserException",
    "Value",
    "ValueRegistry",
    "ValueSchema",
    "ValueSchemaValue",
    "ValueType",
    "Workflow",
    "WorkflowEngine",
    "WorkflowErrors",
    "WorkflowValue",
]



# workflow_engine/core/edge.py

from typing import Self

from ..utils.immutable import ImmutableBaseModel
from .node import Node
from .values import Value


class Edge(ImmutableBaseModel):
    """
    An edge connects the output of source node to the input of a target node.
    """

    source_id: str
    source_key: str
    target_id: str
    target_key: str

    @classmethod
    def from_nodes(
        cls,
        *,
        source: Node,
        source_key: str,
        target: Node,
        target_key: str,
    ) -> Self:
        """
        Self-validating factory method.
        """
        edge = cls(
            source_id=source.id,
            source_key=source_key,
            target_id=target.id,
            target_key=target_key,
        )
        edge.validate_types(source, target)
        return edge

    def validate_types(self, source: Node, target: Node):
        if self.source_key not in source.output_fields:
            raise ValueError(
                f"Source node {source.id} does not have a {self.source_key} field"
            )

        if self.target_key not in target.input_fields:
            raise ValueError(
                f"Target node {target.id} does not have a {self.target_key} field"
            )

        source_output_type, _ = source.output_fields[self.source_key]
        assert issubclass(source_output_type, Value)
        target_input_type, _ = target.input_fields[self.target_key]
        assert issubclass(target_input_type, Value)

        if not source_output_type.can_cast_to(target_input_type):
            raise TypeError(
                f"Edge from {source.id}.{self.source_key} to {target.id}.{self.target_key} has invalid types: {source_output_type} is not assignable to {target_input_type}"
            )

    def with_namespace(self, namespace: str) -> Self:
        return self.model_update(
            source_id=f"{namespace}/{self.source_id}",
            target_id=f"{namespace}/{self.target_id}",
        )


class SynchronizationEdge(ImmutableBaseModel):
    """
    An edge that carries no information, but indicates that the target node must
    run after the source node finishes.
    """

    source_id: str
    target_id: str


__all__ = [
    "Edge",
    "SynchronizationEdge",
]
